level_mod = {
    70: {"lvl_main": 292, "lvl_sub": 364, "lvl_div": 900},
    80: {
        "lvl_main": 340,
        "lvl_sub": 380,
        "lvl_div": 1300,
        "atk_mod": 165,
        "atm_mod_tank": 115,
    },
    90: {
        "lvl_main": 390,
        "lvl_sub": 400,
        "lvl_div": 1900,
        "atk_mod": 195,
        "atk_mod_tank": 156,
    },
    100: {
        "lvl_main": 440,
        "lvl_sub": 420,
        "lvl_div": 2780,
        "atk_mod": 237,
        "atk_mod_tank": 190,
    },
}

pet_defaults = {
    "Summoner": {
        90: {
            "pet_attack_power_scalar": 0.88,
            "pet_attack_power_offset": -48,
            "pet_job_attribute": 100,
            "pet_atk_mod": 195,
        },
        100: {
            "pet_attack_power_scalar": 0.88,
            "pet_attack_power_offset": -64,
            "pet_job_attribute": 100,
            "pet_atk_mod": 237,
        },
    },
    "Astrologian": {
        90: {
            "pet_attack_power_scalar": 1.058,
            "pet_attack_power_offset": 0,
            "pet_job_attribute": 100,
            "pet_atk_mod": 195,
        },
        100: {
            "pet_attack_power_scalar": 1.058,
            "pet_attack_power_offset": 0,
            "pet_job_attribute": 100,
            "pet_atk_mod": 237,
        },
    },
    "DarkKnight": {
        90: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": -18,
            "pet_job_attribute": 100,
            "pet_atk_mod": 195,
        },
        100: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": -18,
            "pet_job_attribute": 100,
            "pet_atk_mod": 237,
        },
    },
    "Ninja": {
        90: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": 0,
            "pet_job_attribute": 100,
            "pet_atk_mod": 195,
        },
        100: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": 0,
            "pet_job_attribute": 100,
            "pet_atk_mod": 237,
        },
    },
    "Machinist": {
        90: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": -61,
            "pet_job_attribute": 100,
            "pet_atk_mod": 195,
        },
        100: {
            "pet_attack_power_scalar": 1.0,
            "pet_attack_power_offset": -61,
            "pet_job_attribute": 100,
            "pet_atk_mod": 237,
        },
    },
}
